ansicolumn t/yamanashi.txt -c80 -P10
ansicolumn t/yamanashi.txt -c80 -P10 -W
ansicolumn t/yamanashi.txt -c80 -P10 -C3
ansicolumn t/yamanashi.txt -c80 -P10 -C3 -D
ansicolumn t/yamanashi.txt -c80 -P10 -C3 -D --runin=4 --runout=4
ansicolumn t/yamanashi.txt t/yamanashi.txt -c80
ansicolumn t/yamanashi-pages.txt -c80 --pages --parallel
ansicolumn t/yamanashi-pages.txt -c80 -P32 -C2 --fillup-str=~
